/*
  CP/M file append program for D88
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* '/' or '\\' for directory separator */
#define DIR_DIVIDE '/'

/* disk parameters */
/* BLS >= SECTOR_SIZE and DSM < 256 */
#define NHEAD 1
#define NCYLINDER 80
#define NSECTOR 16
#define SECTOR_SIZE 256
#define BLS 2048
#define DIR_BLOCK 2
#define OFFSET 2
#define DSM (((NCYLINDER - OFFSET) * NHEAD * NSECTOR * SECTOR_SIZE) / BLS - 1)
#define FNAME_LENGTH 8
#define SUFFIX_LENGTH 3

/* D88 parameters and macro */
#define HEADER_SIZE (32 + 2 * 82 * 4)
#define D88_SIZE (HEADER_SIZE + NCYLINDER * NHEAD * NSECTOR * (16 + SECTOR_SIZE))
#define ADDRESS(a,b,c) (track_table[a][b]+((c)-1)*(16+SECTOR_SIZE)+16)

void next(int *, int *, int *);

int main(int argc, char *argv[])
{
  int iarg;
  int i, j, k;
  long skip, size;
  int block, blsize, logisize;
  int cylinder, head, sector;
  int dcylinder, dhead, dsector, dirpos;
  int use[DSM];
  long track_table[82][2];
  unsigned char d88[D88_SIZE], data[(DSM+1)*BLS];
  char fname[FNAME_LENGTH+1];
  char suffix[SUFFIX_LENGTH+1];
  FILE *fp;

  if (argc < 3){
    printf("usage: cpmcat input.d88 input1.bin [input2.bin ...] output.d88\n");
    exit(1);
  }

  fp = fopen(argv[1], "rb");  /* b for Windows */
  if (fp == NULL){
    printf("cannot open %s\n", argv[1]);
    exit(1);
  }
  fread(d88, sizeof(d88[0]), sizeof(d88)/sizeof(d88[0]), fp);
  fclose(fp);

  for (i = 0; i < sizeof(use) / sizeof(use[0]); i++){
    if (i < DIR_BLOCK){
      use[i] = 1;
    }else{
      use[i] = 0;
    }
  }

  skip = 32;
  for (cylinder = 0; cylinder < 82; cylinder++){
    for (head = 0; head < 2; head++){
      track_table[cylinder][head]
	= d88[skip]
	+ (d88[skip+1]<<8)
	+ (d88[skip+2]<<16)
	+ (d88[skip+3]<<24);
      skip += 4;
    }
  }

  dirpos = -1;
  block = 0;
  for (cylinder = 0; cylinder < NCYLINDER; cylinder++){
    for (head = 0; head < NHEAD; head++){
      for (sector = 1; sector <= NSECTOR; sector++){
	skip = ADDRESS(cylinder, head, sector);
	if (cylinder >= OFFSET && block < DIR_BLOCK){
	  for (i = 0; i < SECTOR_SIZE; i += 32){
	    if (d88[skip] == 0xe5){
	      if(dirpos == -1){
		dcylinder = cylinder;
		dhead = head;
		dsector = sector;
		dirpos = i;
	      }
	      skip += 32;
	    }else{
	      skip += 16;
	      for (j = 0; j < 16; j++){
		if (d88[skip]){
		  use[d88[skip]] = 1;
		}
		skip++;
	      }
	    }
	  }
	  if (sector * SECTOR_SIZE % BLS == 0){
	    block++;
	  }
	}
      }
    }
  }

  block = DIR_BLOCK;
  for (iarg = 2; iarg < argc - 1; iarg++){
    fp = fopen(argv[iarg], "rb");  /* b for Windows */
    if (fp == NULL){
      printf("cannot open %s\n", argv[iarg]);
      exit(1);
    }

    memset(data, 0, sizeof(data));
    size = fread(data, sizeof(data[0]), sizeof(data)/sizeof(data[0]), fp);
    fclose(fp);
    logisize = (size - 1) / 128 + 1;
    blsize = (size - 1) / BLS + 1;

    for (i = strlen(argv[iarg]); i >= 0; i--){
      if (argv[iarg][i] == DIR_DIVIDE){
	break;
      }
    }
    strncpy(fname, "        ", sizeof(fname));
    strncpy(suffix, "   ", sizeof(suffix));

    for (j = i + 1; j < i + 1 + strlen(fname); j++){
      if (argv[iarg][j] == '.'){
	break;
      }
      fname[j-i-1] = toupper((int)argv[iarg][j]);
    }

    for (i = strlen(argv[iarg]); i > 0; i--){
      if (argv[iarg][i] == '.'){
	for(j = 0; j < 3 && i+1 < strlen(argv[iarg]); j++){
	  suffix[j] = toupper((int)argv[iarg][i+1]);
	  i++;
	}
	break;
      }
    }

    for (i = 0; i < (blsize - 1) / 16 + 1; i++){
      skip = ADDRESS(dcylinder, dhead, dsector) + dirpos;
      while(d88[skip] != 0xe5){
	dirpos += 32;
	if (dirpos == SECTOR_SIZE){
	  dirpos = 0;
	  next(&dcylinder, &dhead, &dsector);
	}
	skip = ADDRESS(dcylinder, dhead, dsector) + dirpos;
      }
      d88[skip] = 0;
      memcpy(&d88[skip + 1], fname, FNAME_LENGTH);
      memcpy(&d88[skip + 9], suffix, SUFFIX_LENGTH);
      d88[skip + 12] = i * 16 * BLS / 128 / 128;
      d88[skip + 13] = 0;
      d88[skip + 14] = 0;
      memset(&d88[skip + 16], 0, 16);

      for (j = 0; j < 16 && logisize > 0; j++){
	while(use[block]){
	  block++;
	  if (block >= DSM){
	    printf("disk no space\n");
	    exit(1);
	  }
	}
	cylinder = OFFSET + block * BLS / (NHEAD * NSECTOR * SECTOR_SIZE);
	head = block * BLS / (NSECTOR * SECTOR_SIZE) % NHEAD;
	sector = (block * BLS % (NHEAD * NSECTOR * SECTOR_SIZE)) / SECTOR_SIZE + 1;

	for (k = 0; k < BLS / SECTOR_SIZE; k++){
	  memcpy(&d88[ADDRESS(cylinder, head, sector)],
		 &data[(i * 16 + j) * BLS + k * SECTOR_SIZE], SECTOR_SIZE);
	  next(&cylinder, &head, &sector);
	}
	d88[skip + 16 + j] = block;
	use[block] = 1;
	logisize -= BLS / 128;
      }
      if (logisize > 0){
	d88[skip + 12] += 16 * BLS / 128 / 128 - 1;
	d88[skip + 15] = 128;
      }else{
	d88[skip + 12] += (logisize + j * BLS / 128) / 128;
	d88[skip + 15] = (logisize + j * BLS / 128) % 128;
      }
    }  /* for (i = 0; i < blsize / 16; i ++){ */
  }  /* for (iarg = 2; iarg < argc - 1; iarg++){ */
  fp = fopen(argv[argc - 1], "wb");  /* b for Windows */
  if (fp == NULL){
    printf("cannot open %s\n", argv[argc - 1]);
    exit(1);
  }
  fwrite(d88, sizeof(d88[0]), sizeof(d88)/sizeof(d88[0]), fp);
  fclose(fp);

  return 0;
}

void next(int *cylinder, int *head, int *sector)
{
  (*sector)++;
  if (*sector > NSECTOR){
    (*sector) = 1;
    (*head)++;
    if (*head >= NHEAD){
      (*head) = 0;
      (*cylinder)++;
    }
  }
}
