/*
  PC-6001mkIISR/6601SR CP/M disk image generating program, 2012/12/29
  mybootsr.bin, mybiossr.bin cpm.bin width.com -> cpm.d88
*/

#include <stdio.h>
#include <stdlib.h>

#define HEADER_SIZE 0x2b0
#define NHEAD 1
#define NTRACK 80
#define NSECTOR 16
#define SECTOR_SIZE 256
#define TRACK_DENSITY 2
#define OFF 2
#define BLS 2048
#define DIR_BLOCK 2

void fput2(int, FILE *);
void fput4(long, FILE *);
void header(FILE *);
void secthead(int, int, int, int, FILE *);
void copy(const char *, int *, int *, int *, int *, FILE *);

int main()
{
  FILE *fp;
  int c, h, r, n;
  int block = -1;
  int i;

  fp = fopen ("cpm.d88", "wb");  /* b for Windows */
  if (fp == NULL){
    printf("cannot open output file\n");
    exit(1);
  }

  header(fp);

  c = 0;  /* track */
  h = 0;  /* head */
  r = 1;  /* sector */
  n = SECTOR_SIZE / 256;
  for (i = 0; n > 0; i++){
    n = n / 2;
  }
  n = i;

  copy("mybootsr.bin", &c, &h, &r, &n, fp);
  copy("cpm.bin", &c, &h, &r, &n, fp);
  copy("mybiossr.bin", &c, &h, &r, &n, fp);

  for (; c < OFF; c++){
    for (; h < NHEAD; h++){
      for (; r <= NSECTOR; r++){
	secthead(c, h, r, n, fp);
	for (i = 0; i < SECTOR_SIZE; i++){
	  fputc(0xff, fp);  /* no data */
	}
      }
      r = 1;
    }
    h = 0;
  }

  secthead(c, h, r, n, fp);
  fputc(0x00, fp);
  fwrite("WIDTH   COM", 1, 11, fp);
  fputc(0x00, fp);
  fputc(0x00, fp);
  fputc(0x00, fp);
  fputc(0x01, fp);
  fputc(0x02, fp);
  for (i = 0; i < 15; i++){
    fputc(0x00, fp);
  }

  i = 32;
  for (c = OFF; c < NTRACK; c++){
    for (h = 0; h < NHEAD; h++){
      for (r = 1; r <= NSECTOR; r++){
	if ((r - 1) * SECTOR_SIZE % BLS == 0){
	  block++;
	  if (block == DIR_BLOCK){
	    copy("width.com", &c, &h, &r, &n, fp);
	  }
	}
	if (i == 0){
	  secthead(c, h, r, n, fp);
	}
	for (; i < SECTOR_SIZE; i++){
	  if ( block < DIR_BLOCK){
	    fputc(0xe5, fp);  /* CP/M directory */
	  }else{
	    fputc(0xff, fp);  /* no data */
	  }
	}
	i = 0;
      }
    }
  }

  fclose(fp);
  return 0;
}


void fput2(int data, FILE *fp)
{
  fputc(data&0xff, fp);
  fputc((data>>8)&0xff, fp);
}

void fput4(long data, FILE *fp)
{
  fputc(data&0xff, fp);
  fputc((data>>8)&0xff, fp);
  fputc((data>>16)&0xff, fp);
  fputc((data>>24)&0xff, fp);
}

void header(FILE *fp)
{
  int i, j;
  long size, address;

/* name(17), reserved(9), protect(1) */
  for (i = 0; i < 27; i++){
    fputc(0x00, fp);
  }

/* track density */
  switch (TRACK_DENSITY){
  case 1:
    fputc(0x00, fp);
    break;
  case 2:
    fputc(0x10, fp);
    break;
  default:
    fputc(0x20, fp);
    break;
  }

/* disk size */
  size = HEADER_SIZE + NHEAD * NTRACK * NSECTOR * (SECTOR_SIZE + 16);
  fput4(size, fp);

/* track data table */
  address = HEADER_SIZE;
  for (i = 0; i < 82; i++){  /* track */
    for (j = 0; j < 2; j++){  /* head */
      if (i < NTRACK && j < NHEAD){
	fput4(address, fp);
	address += NSECTOR * (SECTOR_SIZE + 16);
      }else{
	fput4(0, fp);
      }
    }
  }
}

void secthead(int c, int h, int r, int n, FILE *fp)
{
  fputc(c, fp);  /* track */
  fputc(h, fp);  /* head */
  fputc(r, fp);  /* sector */
  fputc(n, fp);  /* sector size=128*2^n */
  fput2(NSECTOR, fp);
  fputc(0, fp);  /* double density */
  fputc(0, fp);  /* not deleted */
  fputc(0, fp);  /* status normal */
  fputc(0, fp);  /* reserved */
  fputc(0, fp);  /* reserved */
  fputc(0, fp);  /* reserved */
  fputc(0, fp);  /* reserved */
  fputc(0, fp);  /* reserved */
  fput2(SECTOR_SIZE, fp);
}

void copy(const char *fname, int *c, int *h, int *r, int *n, FILE *fp_out)
{
  FILE *fp_in;
  char data[SECTOR_SIZE];

  fp_in = fopen(fname, "rb");  /* b for Windows */
  if (fp_in == NULL){
    printf("cannot open %s\n", fname);
    exit(1);
  }

  while((fread(data, sizeof(char), SECTOR_SIZE, fp_in)) == SECTOR_SIZE){
    if (*c > NTRACK-1 || *h > NHEAD-1 || *r > NSECTOR){
      printf("disk full\n");
      exit(1);
    }

    secthead(*c, *h, *r, *n, fp_out);
    fwrite(data, sizeof(char), SECTOR_SIZE, fp_out);
    (*r)++;

    if (*r > NSECTOR){
      (*r) = 1;
      (*h)++;
      if (*h > NHEAD-1){
	(*h) = 0;
	(*c)++;
      }
    }
  }

  fclose(fp_in);
}
